/*
 * Copyright (c) 1982, 1986 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)uipc_domain.c	7.9 (Berkeley) 3/4/91
 */

#include <conf.h>

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/protosw.h>
#include <sys/domain.h>
#include <sys/mbuf.h>
#include <sys/kernel.h>

#include <kern/amiga_includes.h>
#include <kern/uipc_domain_protos.h>

/* --- start moved from sys/domain.h --- */
struct domain *domains = NULL;
/* --- end moved from sys/domain.h --- */

static BOOL domain_initialized = FALSE;

#define	ADDDOMAIN(x)	{ \
	extern struct domain __CONCAT(x,domain); \
	__CONCAT(x,domain.dom_next) = domains; \
	domains = &__CONCAT(x,domain); \
}

BOOL
domaininit ()
{
  register struct domain *dp;
  register struct protosw *pr;

  if (domain_initialized)
    return TRUE;

#undef unix
#ifndef GGTCP
  ADDDOMAIN (unix);
#endif /* GGTCP */
  ADDDOMAIN (route);
#if INET
  ADDDOMAIN (inet);
#endif
#if NS
  ADDDOMAIN (ns);
#endif
#if ISO
  ADDDOMAIN (iso);
#endif
#if RMP
  ADDDOMAIN (rmp);
#endif
#if CCITT
  ADDDOMAIN (ccitt);
#endif
#if NIMP > 0
  ADDDOMAIN (imp);
#endif

  for (dp = domains; dp; dp = dp->dom_next)
    {
      if (dp->dom_init)
	(*dp->dom_init) ();
      for (pr = dp->dom_protosw; pr < dp->dom_protoswNPROTOSW; pr++)
	if (pr->pr_init)
	  (*pr->pr_init) ();
    }
#ifdef GGTCP
  /*
   * No space needed for the link header with SanaII drivers
   */
  max_linkhdr = 0;
#else
  if (max_linkhdr < 16)		/* XXX */
    max_linkhdr = 16;
#endif
  max_hdr = max_linkhdr + max_protohdr;
  max_datalen = MHLEN - max_hdr;

#ifndef GGTCP
  /*
	 * Timeouts are scheduled from amiga_main.c in GGTCP/IP
	 */
  pffasttimo ();
  pfslowtimo ();
#endif
  domain_initialized = TRUE;
  return TRUE;
}

struct protosw *
pffindtype (family, type)
     int family, type;
{
  register struct domain *dp;
  register struct protosw *pr;

  for (dp = domains; dp; dp = dp->dom_next)
    if (dp->dom_family == family)
      goto found;
  return (0);
found:
  for (pr = dp->dom_protosw; pr < dp->dom_protoswNPROTOSW; pr++)
    if (pr->pr_type && pr->pr_type == type)
      return (pr);
  return (0);
}

struct protosw *
pffindproto (family, protocol, type)
     int family, protocol, type;
{
  register struct domain *dp;
  register struct protosw *pr;
  struct protosw *maybe = 0;

  if (family == 0)
    return (0);
  for (dp = domains; dp; dp = dp->dom_next)
    if (dp->dom_family == family)
      goto found;
  return (0);
found:
  for (pr = dp->dom_protosw; pr < dp->dom_protoswNPROTOSW; pr++)
    {
      if ((pr->pr_protocol == protocol) && (pr->pr_type == type))
	return (pr);

      if (type == SOCK_RAW && pr->pr_type == SOCK_RAW &&
	  pr->pr_protocol == 0 && maybe == (struct protosw *) 0)
	maybe = pr;
    }
  return (maybe);
}

void
pfctlinput (cmd, sa)
     int cmd;
     struct sockaddr *sa;
{
  register struct domain *dp;
  register struct protosw *pr;

  for (dp = domains; dp; dp = dp->dom_next)
    for (pr = dp->dom_protosw; pr < dp->dom_protoswNPROTOSW; pr++)
      if (pr->pr_ctlinput)
	(*pr->pr_ctlinput) (cmd, sa, (caddr_t) 0);
}

void
pfslowtimo ()
{
  register struct domain *dp;
  register struct protosw *pr;

  for (dp = domains; dp; dp = dp->dom_next)
    for (pr = dp->dom_protosw; pr < dp->dom_protoswNPROTOSW; pr++)
      if (pr->pr_slowtimo)
	(*pr->pr_slowtimo) ();
#ifndef GGTCP
  /*
	 * Timeouts are scheduled from amiga_main.c in GGTCP/IP
	 */
  timeout (pfslowtimo, (caddr_t) 0, HZ / 2);
#endif /* GGTCP */
}

void
pffasttimo ()
{
  register struct domain *dp;
  register struct protosw *pr;

  for (dp = domains; dp; dp = dp->dom_next)
    for (pr = dp->dom_protosw; pr < dp->dom_protoswNPROTOSW; pr++)
      if (pr->pr_fasttimo)
	(*pr->pr_fasttimo) ();
#ifndef GGTCP
  /*
	 * Timeouts are scheduled from amiga_main.c in GGTCP/IP
	 */
  timeout (pffasttimo, (caddr_t) 0, HZ / 5);
#endif /* GGTCP */
}
